/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/posix/fileutil.c: Miscellaneous file utility functions.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/posix/fileutil.h"
#include "src/sysdep/posix/path_max.h"

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>  // For rename().
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <utime.h>

/*************************************************************************/
/*************************************************************************/

/**
 * SAFE_DLOG:  Call DLOG(), preserving the value of errno across the call.
 */
#ifdef DEBUG
# define SAFE_DLOG(...)  do { \
    int __errno_save = errno; \
    DLOG(__VA_ARGS__);        \
    errno = __errno_save;     \
} while (0)
#else
# define SAFE_DLOG(...)  /*nothing*/
#endif

/*-----------------------------------------------------------------------*/

void *posix_read_file(const char *path, ssize_t *size_ret,
                      unsigned int mem_flags)
{
    PRECOND(path != NULL, errno = EINVAL; return NULL);
    PRECOND(size_ret != NULL, errno = EINVAL; return NULL);

    char *buffer = NULL;

    /* Open the file and get its size. */

    const int fd = open(path, O_RDONLY);
    if (fd < 0) {
        if (errno != ENOENT) {  // Don't complain if it's just a missing file.
            SAFE_DLOG("open(%s) failed: %s", path, strerror(errno));
        }
        goto fail;
    }

    ssize_t size = lseek(fd, 0, SEEK_END);
    if (size < 0) {
        SAFE_DLOG("lseek(%s,0,SEEK_END) failed: %s", path, strerror(errno));
        goto fail_close_fd;
    }
    /* The second seek will always succeed if the first one did. */
    ASSERT(lseek(fd, 0, SEEK_SET) == 0, goto fail_close_fd);

    /* Allocate a buffer for loading the file. */

    buffer = mem_alloc(size > 0 ? size : 1, 0, mem_flags);
    if (!buffer) {
        DLOG("%s: Out of memory (unable to allocate %lld bytes)", path,
             (long long)size);
        errno = ENOMEM;
        goto fail_close_fd;
    }

    /* Read the file contents into the buffer. */

    ssize_t bytes_read = 0;
    while (bytes_read < size) {
        const ssize_t count = read(fd, buffer + bytes_read, size - bytes_read);
        if (count > 0) {
            bytes_read += count;
        } else {
            SAFE_DLOG("read(%s) failed: %s", path,
                      count < 0 ? strerror(errno) : "Short read");
            break;
        }
    }
    if (bytes_read != size) {
        goto fail_close_fd;
    }

    /* Close the file, and return the buffer and its size. */

    close(fd);
    *size_ret = size;
    errno = 0;
    return buffer;

    /* Error cases jump down here. */

  fail_close_fd:
    {
        int errno_save = errno;
        close(fd);
        errno = errno_save;
    }
  fail:
    mem_free(buffer);
    return NULL;
}

/*-----------------------------------------------------------------------*/

int posix_write_file(const char *path, const void *data, ssize_t size)
{
    PRECOND(path != NULL, errno = EINVAL; return 0);
    PRECOND(data != NULL, errno = EINVAL; return 0);
    PRECOND(size >= 0, errno = EINVAL; return 0);

    /* Generate a temporary filename to use for writing, so we don't
     * destroy the original if a write error occurs. */

    char temppath[PATH_MAX];
    const unsigned int temppath_len =
        strformat(temppath, sizeof(temppath), "%s~", path);
    if (temppath_len >= sizeof(temppath)) {
        SAFE_DLOG("Buffer overflow generating temporary pathname for %s",
                  path);
        goto fail;
    }

    /* Create any necessary parent directories. */

    char *s = temppath+1;
    while (s += strcspn(s, "/"), *s != 0) {
        *s = 0;
        if (access(temppath, F_OK) != 0) {
            if (mkdir(temppath, 0777) != 0) {
                SAFE_DLOG("Failed to create parent directory %s of %s: %s",
                          temppath, path, strerror(errno));
                goto fail;
            }
        }
        *s++ = '/';
    }

    /* Write the data to the temporary file. */

    int fd = open(temppath, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (fd < 0) {
        SAFE_DLOG("open(%s) failed: %s", temppath, strerror(errno));
        goto fail;
    }

    ssize_t bytes_written = 0;
    while (bytes_written < size) {
        const ssize_t count =
            write(fd, (const char *)data + bytes_written, size - bytes_written);
        if (count > 0) {
            bytes_written += count;
        } else {
            SAFE_DLOG("write(%s) failed: %s", temppath,
                      count < 0 ? strerror(errno) : "Short write");
            goto fail_unlink;
        }
    }

    const int close_res = close(fd);
    fd = -1;
    if (close_res != 0) {
        SAFE_DLOG("close(%s) failed: %s", temppath, strerror(errno));
        goto fail_unlink;
    }

    /* Rename the temporary file to the final filename.  (POSIX specifies
     * that the rename operation is atomic, so if the rename fails, the
     * original file will not be lost.) */

    if (rename(temppath, path) != 0) {
        SAFE_DLOG("rename(%s, %s) failed: %s", temppath, path,
                  strerror(errno));
        goto fail_unlink;
    }

    /* Success! */

    errno = 0;
    return 1;

    /* Error cases jump down here. */

  fail_unlink:
    {
        int errno_save = errno;
        close(fd);
        unlink(temppath);
        errno = errno_save;
    }
  fail:
    return 0;
}

/*-----------------------------------------------------------------------*/

int posix_copy_file(const char *from, const char *to,
                    int preserve_times, ssize_t buffer_size)
{
    PRECOND(from != NULL, errno = EINVAL; return 0);
    PRECOND(to != NULL, errno = EINVAL; return 0);

    if (buffer_size == 0) {
        buffer_size = 65536;  // Default buffer size.
    }

    /* Retrieve the source file's timestamps if copying is requested.  (We
     * do this first (1) so we don't update the atime before we read it,
     * and (2) so if the source and destination happen to be the same, we
     * still use the correct timestamps.) */

    struct utimbuf ut;
    if (preserve_times) {
        struct stat st;
        if (stat(from, &st) != 0) {
            SAFE_DLOG("stat(%s) failed: %s", from, strerror(errno));
            goto fail;
        }
        ut.actime = st.st_atime;
        ut.modtime = st.st_mtime;
    }

    /* Allocate a temporary buffer for copying. */

    char *buffer = mem_alloc(buffer_size, 0, MEM_ALLOC_TEMP);
    if (!buffer) {
        DLOG("Out of memory (unable to allocate %lld bytes)",
             (long long)buffer_size);
        errno = ENOMEM;
        goto fail;
    }

    /* Open the source file. */

    int from_fd = open(from, O_RDONLY);
    if (from_fd < 0) {
        if (errno != ENOENT) {  // Don't complain if it's just a missing file.
            SAFE_DLOG("open(%s) failed: %s", from, strerror(errno));
        }
        goto fail_free_buffer;
    }

    /* Generate a temporary filename to use for writing and open the
     * temporary file, creating any necessary parent directories. */

    char temppath[PATH_MAX];
    const unsigned int temppath_len =
        strformat(temppath, sizeof(temppath), "%s~", to);
    if (temppath_len >= sizeof(temppath)) {
        SAFE_DLOG("Buffer overflow generating temporary pathname for %s", to);
        goto fail_close_from_fd;
    }

    char *s = temppath+1;
    while (s += strcspn(s, "/"), *s != 0) {
        *s = 0;
        if (access(temppath, F_OK) != 0) {
            if (mkdir(temppath, 0777) != 0) {
                DLOG("Failed to create parent directory %s of %s: %s",
                     temppath, to, strerror(errno));
                goto fail_close_from_fd;
            }
        }
        *s++ = '/';
    }

    const int to_fd = open(temppath, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (to_fd < 0) {
        SAFE_DLOG("open(%s) failed: %s", temppath, strerror(errno));
        goto fail_close_from_fd;
    }

    /* Copy the file contents. */

    ssize_t bytes_read;
    for (;;) {
        bytes_read = read(from_fd, buffer, buffer_size);
        if (bytes_read <= 0) {
            if (bytes_read < 0) {
                SAFE_DLOG("read(%s) failed: %s", from,
                          bytes_read < 0 ? strerror(errno) : "Short read");
            }
            break;
        }
        const ssize_t bytes_written = write(to_fd, buffer, bytes_read);
        if (bytes_written != bytes_read) {
            SAFE_DLOG("write(%s) failed: %s", to,
                      bytes_written < 0 ? strerror(errno) : "Short write");
            break;
        }
    }
    if (bytes_read != 0) {
        goto fail_unlink;
    }

    /* Free the copy buffer and close the files. */

    mem_free(buffer);
    buffer = NULL;
    close(from_fd);
    /* close() failure on the input file is meaningless to us. */
    from_fd = -1;
    const int close_res = close(to_fd);
    if (close_res != 0) {
        SAFE_DLOG("close(%s) failed: %s", temppath, strerror(errno));
        goto fail_unlink;
    }

    /* Update the output file's timestamps, if requested. */

    if (preserve_times) {
        if (utime(temppath, &ut) != 0) {
            SAFE_DLOG("utime(%s) failed: %s", temppath, strerror(errno));
            goto fail_unlink;
        }
    }

    /* Move the temporary output file to its final name. */

    if (rename(temppath, to) != 0) {
        SAFE_DLOG("rename(%s, %s) failed: %s", temppath, to, strerror(errno));
        goto fail_unlink;
    }

    /* Success! */

    errno = 0;
    return 1;

    /* Error cases jump down here. */

  fail_unlink:
    {
        int errno_save = errno;
        unlink(temppath);
        close(to_fd);
        errno = errno_save;
    }
  fail_close_from_fd:
    if (from_fd >= 0) {
        int errno_save = errno;
        close(from_fd);
        errno = errno_save;
    }
  fail_free_buffer:
    mem_free(buffer);
  fail:
    return 0;
}

/*-----------------------------------------------------------------------*/

int posix_rmdir_r(const char *path)
{
    int ok = 1;

    DIR *d = opendir(path);
    if (!d) {
        SAFE_DLOG("Failed to scan directory %s: %s", path, strerror(errno));
        return 0;
    }

    int errno_save = 0;
    struct dirent *de;
    while ((de = readdir(d)) != NULL) {
        if (strcmp(de->d_name, ".") != 0 && strcmp(de->d_name, "..") != 0) {
            char temppath[PATH_MAX];
            if (!strformat_check(temppath, sizeof(temppath),
                                 "%s/%s", path, de->d_name)) {
                DLOG("Failed to remove %s/%s: Buffer overflow",
                     path, de->d_name);
                errno_save = ENAMETOOLONG;
                ok = 0;
            } else {
                struct stat st;
                if (lstat(temppath, &st) == 0 && S_ISDIR(st.st_mode)) {
                    if (!posix_rmdir_r(temppath)) {
                        errno_save = errno;
                        ok = 0;
                    }
                } else if (unlink(temppath) != 0) {
                    SAFE_DLOG("Failed to remove %s: %s", temppath,
                              strerror(errno));
                    errno_save = errno;
                    ok = 0;
                }
            }
        }
    }
    closedir(d);
    errno = errno_save;

    if (ok && rmdir(path) != 0) {
        SAFE_DLOG("Failed to remove directory %s: %s", path, strerror(errno));
        ok = 0;
    }

    return ok;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
