/*
 * System Interface Library for games
 * Copyright (c) 2011 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/sysdep/windows/userdata.c: Tests for the Windows implementation
 * of the user data access functions.
 */

#ifdef INCLUDE_TESTS  // To the end of the file.

#include "src/base.h"
#include "src/graphics.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/test.h"
#include "src/sysdep/windows/internal.h"
#include "src/sysdep/windows/userdata.h"
#include "src/test/base.h"
#include "src/test/graphics/internal.h"
#include "src/test/userdata.h"
#include "src/texture.h"
#include "src/userdata.h"

/*************************************************************************/
/*************************** Helper functions ****************************/
/*************************************************************************/

/**
 * remove_dir:  Remove a directory and all files contained therein.
 * Subdirectories are not removed, and the function will fail if any
 * subdirectories are present.
 *
 * [Parameters]
 *     path: Pathname of directory to remove.
 * [Return value]
 *     True on success or if the directory does not exist, false on error.
 */
static int remove_dir(const char *path)
{
    PRECOND(path != NULL);
    SysDir *dir = sys_dir_open(path);
    if (dir) {
        const char *name;
        while ((name = sys_dir_read(dir)) != NULL) {
            /* While MSDN is silent on the subject, it seems to be
             * generally regarded as safe to delete files in a
             * FindNextFile() loop; it is also safe in Wine after
             * http://bugs.winehq.org/show_bug.cgi?id=31987 was fixed. */
            char file_path[MAX_PATH+1];
            ASSERT(strformat_check(file_path, sizeof(file_path),
                                   "%s\\%s", path, name));
            if (UNLIKELY(!DeleteFile(file_path))) {
                DLOG("Failed to delete %s: %s", file_path,
                     windows_strerror(GetLastError()));
                return 0;
            }
        }
        sys_dir_close(dir);
        if (UNLIKELY(!RemoveDirectory(path))) {
            DLOG("Failed to delete %s: %s", path,
                 windows_strerror(GetLastError()));
            return 0;
        }
    }
    return 1;
}

/*-----------------------------------------------------------------------*/

/**
 * windows_init:  Initialization for userdata tests specific to the Windows
 * implementation.
 */
static int windows_init(void)
{
    static char tempdir[MAX_PATH+2];
    const unsigned int templen = GetTempPath(sizeof(tempdir), tempdir);
    if (UNLIKELY(templen == 0)) {
        DLOG("GetTempPath() failed: %s", windows_strerror(GetLastError()));
        return 0;
    }
    ASSERT(templen <= MAX_PATH+1);  // Guaranteed by API.
    /* Make sure the final path for all files will fit in MAX_PATH
     * characters: <tempdir>\SIL-1234567890\screenshots\screen0.png */
    if (templen > MAX_PATH-40) {
        DLOG("Temporary directory pathname %s is too long", tempdir);
        return 0;
    }
    ASSERT(strformat_check(tempdir+templen, sizeof(tempdir)-templen,
                           "SIL-%u\\", (int)GetCurrentProcessId()));
    TEST_windows_userdata_path = tempdir;
    sys_test_userdata_use_live_routines = 1;
    return 1;
}

/*-----------------------------------------------------------------------*/

/**
 * windows_cleanup:  Cleanup for userdata tests specific to the Windows
 * implementation.
 */
static int windows_cleanup(void)
{
    sys_test_userdata_use_live_routines = 0;

    char save_path[MAX_PATH+1];
    ASSERT(strformat_check(save_path, sizeof(save_path), "%ssave",
                           TEST_windows_userdata_path));
    char screenshot_path[MAX_PATH+1];
    ASSERT(strformat_check(screenshot_path, sizeof(screenshot_path),
                           "%sscreenshots", TEST_windows_userdata_path));
    CHECK_TRUE(remove_dir(save_path));
    CHECK_TRUE(remove_dir(screenshot_path));
    CHECK_TRUE(remove_dir(TEST_windows_userdata_path));

    TEST_windows_userdata_path = NULL;
    TEST_windows_userdata_unregister_stats();
    return 1;
}

/*-----------------------------------------------------------------------*/

#ifdef SIL_UTILITY_INCLUDE_PNG

/**
 * get_screenshot:  System-specific function to get saved
 * screenshot data for the stub implementation of the user data access
 * functions.
 */
static void *get_screenshot(int index, int *width_ret, int *height_ret)
{
    char png_path[PATH_MAX+1];
    ASSERT(strformat_check(png_path, sizeof(png_path),
                           "%sscreenshots\\screen%d.png",
                           TEST_windows_userdata_path, index));
    SysFile *fh = sys_file_open(png_path);
    if (!fh) {
        return NULL;
    }
    const int png_size = sys_file_size(fh);
    void *png_data;
    ASSERT(png_data = mem_alloc(png_size, 0, MEM_ALLOC_TEMP));
    if (sys_file_read(fh, png_data, png_size) != png_size) {
        DLOG("Error reading from %s", png_path);
        mem_free(png_data);
        sys_file_close(fh);
        return NULL;
    }
    sys_file_close(fh);

    int image;
    CHECK_TRUE(image = texture_parse(png_data, png_size, 0, 0, 1));
    const void *image_pixels = texture_lock_readonly(image);
    if (!image_pixels) {
        texture_destroy(image);
        return NULL;
    }
    *width_ret = texture_width(image);
    *height_ret = texture_height(image);

    void *copy;
    ASSERT(copy = mem_alloc((*width_ret) * (*height_ret) * 4, 0, 0));
    memcpy(copy, image_pixels, (*width_ret) * (*height_ret) * 4);
    texture_destroy(image);

    return copy;
}

#endif  // SIL_UTILITY_INCLUDE_PNG

/*************************************************************************/
/***************** Test runner and init/cleanup routines *****************/
/*************************************************************************/

static int do_test_windows_userdata(void);
int test_windows_userdata(void)
{
    return run_tests_in_window(do_test_windows_userdata);
}

DEFINE_GENERIC_TEST_RUNNER(do_test_windows_userdata)

TEST_INIT(init)
{
    TEST_userdata_reset();
    userdata_set_program_name("test", "Userdata Test");
    CHECK_TRUE(windows_init());
    return 1;
}

TEST_CLEANUP(cleanup)
{
    if (!sys_test_userdata_use_live_routines) {
        return 1;  // Outer call for nested tests.
    }

    graphics_flush_resources();
    TEST_userdata_reset();
    CHECK_TRUE(windows_cleanup());
    return 1;
}

/*************************************************************************/
/***************************** Test routines *****************************/
/*************************************************************************/

TEST(test_common)
{
    /* If we try to make nested calls to init() and cleanup(), path buffers
     * will be overwritten and things will get generally confused, so we
     * call cleanup() for this test now.  The test at the top of cleanup()
     * will ensure that it doesn't get run when this test returns. */
    ASSERT(cleanup());

    /* Re-run the common userdata tests using the Windows implementation. */
    DLOG("Running common userdata tests for Windows...");
    if (!run_userdata_tests(
            1, windows_init, windows_cleanup,
#ifdef SIL_UTILITY_INCLUDE_PNG
            get_screenshot,
#else
            NULL,
#endif
            NULL))
    {
        FAIL("Preceding failure(s) occurred while testing Windows userdata"
             " functions");
    }

    DLOG("Common userdata tests for Windows succeeded.");
    return 1;
}

/*************************************************************************/
/*************************************************************************/

#endif  // INCLUDE_TESTS

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
